/* -*-c-*-
 *
 * The SHA256 hash function (compact edition)
 *
 * (c) 2020 Mark Wooding
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Runlisp, a tool for invoking Common Lisp scripts.
 *
 * Runlisp is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 3 of the License, or (at your
 * option) any later version.
 *
 * Runlisp is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Runlisp.  If not, see <https://www.gnu.org/licenses/>.
 */

/*----- Header files ------------------------------------------------------*/

#include <string.h>

#include "sha256.h"

/*----- Preliminary definitions -------------------------------------------*/

/* The initial values of the state variables.  These are in reverse order --
 * see the note in `compress'.
 */
static const u32 iv[8] = {
  0x5be0cd19, 0x1f83d9ab, 0x9b05688c, 0x510e527f,
  0xa54ff53a, 0x3c6ef372, 0xbb67ae85, 0x6a09e667
};

/* The round constants. */
static const u32 rc[64] = {
  0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5,
  0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
  0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
  0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
  0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc,
  0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
  0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7,
  0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
  0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
  0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
  0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3,
  0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
  0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5,
  0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
  0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
  0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2
};

/* Standard bithacking operations on 32-bit words.
 *
 * Note that this code assumes that a `u32' is /at least/ 32 bits wide, but
 * may be longer, so we must do some work to keep cruft in the top bits from
 * messing things up.
 */
#define M32 0xffffffff
#define LSL32(x, n) ((x) << ((n)))
#define LSR32(x, n) (((x)&M32) >> ((n)))
#define ROR32(x, n) (LSL32((x), 32 - (n)) | LSR32((x), (n)))

/* Reading and writing 32-bit words. */
#define LOAD32_B(p)							\
  (((u32)(((const unsigned char *)(p))[0]&0xff) << 24) |		\
   ((u32)(((const unsigned char *)(p))[1]&0xff) << 16) |		\
   ((u32)(((const unsigned char *)(p))[2]&0xff) <<  8) |		\
   ((u32)(((const unsigned char *)(p))[3]&0xff) <<  0))
#define STORE32_B(p, x) do {						\
  (void)sizeof(memmove((p), (p), 1));					\
  ((unsigned char *)(p))[0] = ((x) >> 24)&0xff;				\
  ((unsigned char *)(p))[1] = ((x) >> 16)&0xff;				\
  ((unsigned char *)(p))[2] = ((x) >>  8)&0xff;				\
  ((unsigned char *)(p))[3] = ((x) >>  0)&0xff;				\
} while (0)

/* SHA256's balanced ternary operators. */
#define CH(x, y, z) (((x)&(y)) | (~(x)&(z)))
#define MAJ(x, y, z) (((x)&(y)) | ((y)&(z)) | ((z)&(x)))

/* The SHA256 Σ and σ functions. */
#define S0(x) (ROR32((x),  2) ^ ROR32((x), 13) ^ ROR32((x), 22))
#define S1(x) (ROR32((x),  6) ^ ROR32((x), 11) ^ ROR32((x), 25))
#define s0(x) (ROR32((x),  7) ^ ROR32((x), 18) ^ LSR32((x),  3))
#define s1(x) (ROR32((x), 17) ^ ROR32((x), 19) ^ LSR32((x), 10))

/*----- Main code ---------------------------------------------------------*/

/* Compress a 64-byte buffer at P, updating the hash state S. */
static void compress(struct sha256_state *s, const unsigned char *p)
{
  u32 t, u, a[8], m[16];
  const u32 *r = rc;
  size_t i;

  /* This is a mostly straightforward implementation of the specification, as
   * a rolled-up loop, one iteration per round.  The only wrinkle is that the
   * vector of state variables, conventionally named a, b, ..., h, are
   * maintained in our state structure in reverse order, so h is in S->a[0],
   * b is in S->a[6], and a is in S->a[7].  We do this so that we advance
   * through our vector in the correct direction from round to round: this
   * avoids making the indexing arithmetic too complicated.
   */

  /* Move the state and message data into our internal vectors. */
  for (i = 0; i < 8; i++) a[i] = s->a[i];
  for (i = 0; i < 16; i++, p += 4) m[i] = LOAD32_B(p);

  /* Perform 64 rounds of update.  Update the message schedule as we go.  The
   * last 16 rounds of message-schedule update are pointless: doing the
   * message-schedule update conditionally would make the loop messier, and
   * running the message schedule separately would add a second loop and
   * require more intermediate storage.
   */
  for (i = 0; i < 64; i++) {
#define A(j) (a[(i + (j))%8])
#define M(j) (m[(i + (j))%16])
    t = A(0) + S1(A(3)) + CH(A(3), A(2), A(1)) + M(0) + *r++;
    u = S0(A(7)) + MAJ(A(7), A(6), A(5));
    A(4) += t; A(0) = t + u;
    M(0) += s1(M(14)) + M(9) + s0(M(1));
#undef A
#undef M
  }

  /* Write out the updated state. */
  for (i = 0; i < 8; i++) s->a[i] += a[i];
}

/* Initialize the hash state S. */
void sha256_init(struct sha256_state *s)
  { size_t i; s->n = s->nblk = 0; for (i = 0; i < 8; i++) s->a[i] = iv[i]; }

/* Append SZ bytes of data starting at M to the hash state S. */
void sha256_hash(struct sha256_state *s, const void *m, size_t sz)
{
  const unsigned char *p = m;
  size_t r = SHA256_BLKSZ - s->n;

  /* Feed the input data into the hash function.  Our buffer-management
   * policy is to empty the buffer by calling the compression function as
   * soon as the buffer fills completely.
   */
  if (sz < r) {
    /* The whole input will fit into the buffer, with space to spare.  We
     * just copy it in and update the occupancy counter.
     */

    memcpy(s->buf + s->n, p, sz);
    s->n += sz;
  } else {
    /* We're going to fill the buffer at least once. */

    /* If the buffer contains any data already then copy the initial portion
     * of the new input chunk into the buffer and compress it there.
     * Otherwise, if the buffer is entirely empty, then we can compress the
     * initial block from the input directly.
     */
    if (!s->n) { compress(s, p); p += SHA256_BLKSZ; sz -= SHA256_BLKSZ; }
    else { memcpy(s->buf + s->n, p, r); compress(s, s->buf); p += r; sz -= r; }
    s->nblk++;

    /* Continue compressing complete blocks from the input while enough
     * material remains.
     */
    while (sz >= SHA256_BLKSZ)
      { compress(s, p); s->nblk++; p += SHA256_BLKSZ; sz -= SHA256_BLKSZ; }

    /* Copy the tail end into the buffer and record how much there is. */
    s->n = sz; if (sz) memcpy(s->buf, p, sz);
  }
}

/* Write the final hash of state S to buffer H. */
void sha256_done(struct sha256_state *s, unsigned char *h)
{
  size_t i, n, r;
  u32 lo, hi;

  /* Add the end-of-data marker to the buffer.  There must be at least one
   * byte spare, or we'd have compressed already.
   */
  n = s->n; s->buf[n++] = 0x80; r = SHA256_BLKSZ - n;

  /* If there's enough space for the message length, then fill the gap
   * between with zeros.  Otherwise, fill the whole of the remaining space,
   * compress, and then refill the initial portion of the buffer.  Either
   * way, after this, there's just eight bytes left at the end of the buffer,
   * into which we can drop the length.
   */
  if (r >= 8)
    memset(s->buf + n, 0, r - 8);
  else {
    if (r) memset(s->buf + n, 0, r);
    compress(s, s->buf);
    memset(s->buf, 0, SHA256_BLKSZ - 8);
  }

  /* Convert the length into two 32-bit halves measuring the total input
   * length in bits, and run the compression function one last time.  There
   * can be no carry, since S->n is always less than 64.
   */
  lo = ((s->nblk <<  9) | (s->n <<  3))&M32; hi = (s->nblk >> 23)&M32;
  STORE32_B(s->buf + 56, hi); STORE32_B(s->buf + 60, lo);
  compress(s, s->buf);

  /* Write out the final hash value.  We must compensate here because the
   * state variables are in reverse order.
   */
  for (i = 8; i-- > 0; h += 4) STORE32_B(h, s->a[i]);
}

/*----- That's all, folks -------------------------------------------------*/
